@@ -8,7 +8,7 @@ use Scalar::Util qw(openhandle);
 use Fcntl qw(O_CREAT O_EXCL O_WRONLY);
 use 5.008;
 
-our $VERSION = '1.16';
+our $VERSION = '1.17';
 
 
 has 'confname' => (
@@ -70,6 +70,12 @@ has 'encoding' => (
     isa => Maybe[Str],
 );
 
+has 'newlines' => (
+    is => 'rw',
+    isa => HashRef,
+    default => sub { +{} },
+);
+
 has 'include' => (
     is => 'rw',
     isa => Str,
@@ -165,6 +171,19 @@ sub _read_config {
 
     my $c = do {local $/; <$fh>};
 
+    my $newlines = "\n";
+    if ($c =~ m/\r\n/) {
+        # Convert from DOS; `git` applies this on read always, and
+        # simply mangles files on write.
+        $newlines = "\r\n";
+        $c =~ s/\r\n/\n/g;
+    } elsif ($c !~ /\n/ and $c =~ /\r/) {
+        # Best-guess convert from Mac.
+        $newlines = "\r";
+        $c =~ s/\r/\n/g;
+    }
+    $self->newlines->{$filename} = $newlines;
+
     $c =~ s/\n*$/\n/; # Ensure it ends with a newline
 
     return $c;
@@ -675,14 +694,26 @@ sub get_regexp {
         }
         elsif ($args{filter} =~ s/^!//) {
             for (keys %results) {
-                delete $results{$_} if defined $results{$_}
-                    and $results{$_} =~ m/$args{filter}/i;
+                my @values = ref $results{$_} ? @{$results{$_}} : $results{$_};
+                @values = grep { not defined or not m/$args{filter}/i } @values;
+                if (!@values) {
+                    delete $results{$_};
+                }
+                else {
+                    $results{$_} = @values > 1 ? \@values : $values[0];
+                }
             }
         }
         else {
             for (keys %results) {
-                delete $results{$_} if not defined $results{$_}
-                    or $results{$_} !~ m/$args{filter}/i;
+                my @values = ref $results{$_} ? @{$results{$_}} : $results{$_};
+                @values = grep { defined and m/$args{filter}/i } @values;
+                if (!@values) {
+                    delete $results{$_};
+                }
+                else {
+                    $results{$_} = @values > 1 ? \@values : $values[0];
+                }
             }
         }
     }
@@ -706,6 +737,8 @@ sub canonical_case {
     my $self = shift;
     my ($key) = @_;
     my ($section, $subsection, $name) = _split_key($key);
+    die "No section given in key: $key\n" unless $section;
+
     return join( '.',
         grep { defined } (lc $section, $subsection, lc $name),
     );
@@ -1077,6 +1110,8 @@ sub _write_config {
     my $self = shift;
     my($filename, $content) = @_;
 
+    my $newlines = $self->newlines->{$filename} || "\n";
+    $content =~ s/\n/$newlines/g if $newlines ne "\n";
     # allow nested symlinks but only within reason
     my $max_depth = 5;
 
@@ -1902,6 +1937,11 @@ Git just rejects them.
 We don't support NUL-terminating output (the --null flag to
 git-config). Who needs it?
 
+Git only supports reading UNIX- and DOS-style newlines ("\n" and
+"\r\n"), and always uses "\n" when modifying files.  We also support
+reading Mac-style newlines ("\r"), and write updates to files using
+the same newlines as they were read with.
+
 =head1 BUGS
 
 If you find any bugs in this module, report them at:
